import es
import os.path
import AchievementGroup as AchiGroup
import Achievement as Achi

# == ACHIEVEMENTIO BASE CLASS ===========================
class AchievementIO:
    """
    Base class for Achievement IO functionality
    """
    __cache = {}
    debugLevel = 1
    
    def load(self):
        """
        Loads achievement data
        [BASE: Does very simple logging]
        """
        es.log("[MYOA] Achievement data loaded")
        es.dbgmsg(self.debugLevel, "[MYOA] Achievement data loaded.")
    
    def save(self):
        """
        Saves achievement data
        [BASE: Does very simple logging]
        """
        es.log("[MYOA] Achievement data saved")
        es.dbgmsg(self.debugLevel, "[MYOA] Achievement data saved.")
    
    def close(self):
        """
        Closes any open connections with the data
        [BASE: Does very simple logging]
        """
        es.log("[MYOA] Achievement data safely closed")
        es.dbgmsg(self.debugLevel, "[MYOA] Achievement data safely closed.")
    
    def onPlayerConnect(self, steamID):
        """
        Is called when a player is connected
        [BASE: Does nothing]
        """
        pass
    
    def onRoundEnd(self):
        """
        Is called when a round ends
        [BASE: Does nothing]
        """
        pass
    
    def setAchievementGroupProperties(self, groupID, title, description):
        """
        Sets the achievement group properties in the database (only for web-based databases)
        [BASE: Does very simple logging]
        """
        es.log("[MYOA] Updated achievement group properties for achievement group %s" % groupID)
        es.dbgmsg(self.debugLevel, "[MYOA] Updated achievement group properties for achievement group %s" % groupID)
    
    def setAchievementProperties(self, groupID, achievementID, title, description, progress, imgPath):
        """
        Sets the achievement propeties in the database (only for web-based databases)
        [BASE: Does very simple logging]
        """
        es.log("[MYOA] Updated achievement properties for achievement [%s, %s]" % (groupID, achievementID))
        es.dbgmsg(self.debugLevel, "[MYOA] Updated achievement properties for achievement [%s, %s]" % (groupID, achievementID))
    
    def getPlayerProgress(self, groupID, achievementID, steamID):
        """
        Retrieves the value of the progress for a particular player for a particular achievement
        [BASE: Returns the cached value if one found, or None if no cached value can be found]
        """
        if not steamID in self.__cache.keys():
            return None
        if not "Achievements" in self.__cache[steamID].keys():
            return None
        if not groupID in self.__cache[steamID]["Achievements"].keys():
            return None
        if not achievementID in self.__cache[steamID]["Achievements"][groupID].keys():
            return None
        return self.__cache[steamID]["Achievements"][groupID][achievementID]

    def setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly = False):
        """
        Sets the value of the progress for a particular player for a particular achievement
        [BASE: Updates the local cached value]
        """
        if not steamID in self.__cache.keys():
            self.__cache[steamID] = { "Achievements": { groupID: { achievementID: progress } } }
        elif not "Achievements" in self.__cache[steamID].keys():
            self.__cache[steamID]["Achievements"] = { groupID: { achievementID: progress } }
        elif not groupID in self.__cache[steamID]["Achievements"].keys():
            self.__cache[steamID]["Achievements"][groupID] = { achievementID: progress }
        else:
            self.__cache[steamID]["Achievements"][groupID][achievementID] = progress
        
        if not cacheOnly:
            es.log("[MYOA] \"%s<%s>\"[%s, %s] = %d" % (self.getPlayerName(steamID), steamID, groupID, achievementID, int(progress)))
    
    def getPlayerName(self, steamID):
        """
        Retrieves the value of the name for a particular player
        [BASE: Returns the cached value if one found, or None if no cached value can be found]
        """
        if not steamID in self.__cache.keys():
            return None
        if not "Name" in self.__cache[steamID].keys():
            return None
        return self.__cache[steamID]["Name"]
    
    def setPlayerName(self, steamID, name, cacheOnly = False):
        """
        Sets the value of the name for a particular player
        [BASE: Updates the local cached value]
        """
        if not steamID in self.__cache.keys():
            self.__cache[steamID] = { "Name": name }
        else:
            self.__cache[steamID]["Name"] = name
    
    def getPlayerAchievementCount(self, steamID):
        """
        Retrieves the value of the achievement count for a particular player
        [BASE: Returns the cached value if one found, or None if no cached value can be found]
        """
        if not steamID in self.__cache.keys():
            return None
        if not "AchievementCount" in self.__cache[steamID].keys():
            return None
        return self.__cache[steamID]["AchievementCount"]
    
    def setPlayerAchievementCount(self, steamID, count, cacheOnly = False):
        """
        Sets the value of the achievement count for a particular player
        [BASE: Updates the local cached value]
        """
        if not steamID in self.__cache.keys():
            self.__cache[steamID] = { "AchievementCount": name }
        else:
            self.__cache[steamID]["AchievementCount"] = count

        if not cacheOnly:
            es.log("[MYOA] \"%s<%s>\"[TOTAL] = %d" % (self.getPlayerName(steamID), steamID, int(count)))
    
    def onPlayerDisconnect(self, steamID):
        """
        Is called when a player is disconnected
        [BASE: Removes any cached data about the player from the cache dictionary]
        """
        if steamID in self.__cache.keys():
            del self.__cache[steamID]
    
    def onPlayerAchievementEvent(self, steamID, groupID, achievementID, event):
        """
        Is called when a player triggers an achievement event (achieved/un-achieved/reset)
        [BASE: Does nothing]
        """
        pass
            
    def isPlayerCached(self, steamID):
        return steamID in self.__cache.keys() and "Achievements" in self.__cache[steamID].keys()
# =======================================================

# == ACHIEVEMENTIOPICKLE CLASS ==========================
try:
    import cPickle
    
    class AchievementIOPickle(AchievementIO):
        """
        Achievement IO functionality for cPickle
        """
        def __init__(self, achievementFilename, ranksFilename, pairsFilename):
            self.achievementFilename = achievementFilename
            self.ranksFilename = ranksFilename
            self.pairsFilename = pairsFilename
            
            #Create empty dictionaries ready for data
            self.__achievements = {}
            self.__playerPairs = {}
            self.__rankings = {}
            self.load()
                            
        def load(self):
            """
            Loads achievement data from cPickle'd file
            """
            addonPath = es.getAddonPath("myo_achievement")
            #Load player achievement progress data
            if os.path.exists(addonPath + "/" + self.achievementFilename):
                pickleFile = open(addonPath + "/" + self.achievementFilename, "r")
                
                database = cPickle.load(pickleFile)
                #Now check to see if the cPickle was saved using the old structure, if so, update it
                
                oldDatabase = False
                for key in database.keys():
                    if isinstance(database[key], AchiGroup.AchievementGroup):
                        oldDatabase = True
                        for ach in database[key].achievements.values():
                            if isinstance(ach, Achi.Achievement):
                                for uniqueID in ach.playerprogress.keys():
                                    self.setPlayerProgress(database[key].id, ach.id, uniqueID, ach.playerprogress[uniqueID])
                
                if not oldDatabase:
                    self.__achievements = database
                else:
                    es.dbgmsg(self.debugLevel, "[MYOA] cPickle database was saved using old schema. Data has been updated to fit the new schema.")
                            
                pickleFile.close()
            
            #Load player pair/name data
            if os.path.exists(addonPath + "/" + self.pairsFilename):
                pickleFile = open(addonPath + "/" + self.pairsFilename, "r")
                self.__playerPairs = cPickle.load(pickleFile)
                pickleFile.close()

            #Load player rank/achievement count data
            if os.path.exists(addonPath + "/" + self.ranksFilename):
                pickleFile = open(addonPath + "/" + self.ranksFilename, "r")
                self.__rankings = cPickle.load(pickleFile)
                pickleFile.close()
            
            #Call base class
            AchievementIO.load(self)
                        
        def save(self):
            """
            Saves achievement data to cPickle'd file
            """
            addonPath = es.getAddonPath("myo_achievement")
        
            #Save player achievement progress data
            pickleFile = open(addonPath + "/" + self.achievementFilename, "w")
            cPickle.dump(self.__achievements, pickleFile)
            pickleFile.close()
            
            #Save player pair/name data
            pickleFile = open(addonPath + "/" + self.pairsFilename, "w")
            cPickle.dump(self.__playerPairs, pickleFile)
            pickleFile.close()        

            #Save player rank/achievement count data
            pickleFile = open(addonPath + "/" + self.ranksFilename, "w")
            cPickle.dump(self.__rankings, pickleFile)
            pickleFile.close()
            
            #Call base class
            AchievementIO.save(self)
        
        def close(self):
            """
            Does nothing, as no extra work is required to close the files
            """
            #Call base class
            AchievementIO.close(self)
        
        def getPlayerProgress(self, groupID, achievementID, steamID):
            """
            Retrieves the value of the progress for a particular player for a particular achievement
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerProgress(self, groupID, achievementID, steamID)
            if not cachedValue == None:
                return cachedValue
            
            if not groupID in self.__achievements.keys():
                return 0
            
            if not achievementID in self.__achievements[groupID].keys():
                return 0
            
            if not steamID in self.__achievements[groupID][achievementID].keys():
                return 0
            else:
                return int(self.__achievements[groupID][achievmentID][steamID])
                    
        def setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly = False):           
            """
            Sets the value of the progress for a particular player for a particular achievement
            """
            if not cacheOnly:
                if not groupID in self.__achievements.keys():
                    self.__achievements[groupID] = { achievementID: { steamID: progress } }
                elif not achievementID in self.__achievements[groupID].keys():
                    self.__achievements[groupID][achievementID] = { steamID: progress }
                else:
                    self.__achievements[groupID][achievementID][steamID] = int(progress)
              
            #Call base class  
            AchievementIO.setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly)
        
        def getPlayerName(self, steamID):
            """
            Retrieves the value of the name for a particular player
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerName(self, steamID)
            if not cachedValue == None:
                return cachedValue

            if steamID in self.__playerPairs.keys():
                return self.__playerPairs[steamID]
            return ""
        
        def setPlayerName(self, steamID, name, cacheOnly = False):
            """
            Sets the value of the name for a particular player
            """
            if not cacheOnly:
                self.__playerPairs[steamID] = name
            
            #Call base class
            AchievementIO.setPlayerName(self, steamID, name, cacheOnly)
            
        def getPlayerAchievementCount(self, steamID):
            """
            Retrieves the value of the achievement count for a particular player
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerAchievementCount(self, steamID)
            if not cachedValue == None:
                return cachedValue

            if steamID in self.__rankings.keys():
                return int(self.__rankings[steamID])
            return 0
        
        def setPlayerAchievementCount(self, steamID, count, cacheOnly = False):
            """
            Sets the value of the achievement count for a particular player
            """
            if not cacheOnly:
                self.__rankings[steamID] = int(count)
            
            #Call base class
            AchievementIO.setPlayerAchievementCount(self, steamID, count, cacheOnly)
except:    
    es.dbgmsg(self.debugLevel, "[MYOA] Error: Could not load cPickle module. Cannot use cPickle as a storage method.")
# =======================================================

# == ACHIEVEMENTIOSQLITE CLASS ==========================    
try:
    import sqlite3
    
    class AchievementIOSqlite(AchievementIO):
        def __init__(self, filename, onlyRoundEnd = False):
            self.databaseFilename = filename
            self.onlyRoundEnd = onlyRoundEnd
            self.__roundEndBuffer = []
            self.load()
        
        def load(self):
            """
            Loads achievement data from Sqlite database
            """
            database = es.getAddonPath("myo_achievement") + "/" + self.databaseFilename
            self.__db_connection = sqlite3.connect(database)
            self.__db_cursor = self.__db_connection.cursor()            
            self.__db_cursor.execute("CREATE TABLE IF NOT EXISTS achievements_progress (steamid TEXT,groupid TEXT,achievementid TEXT,progress INTEGER DEFAULT '0')")
            self.__db_cursor.execute("CREATE TABLE IF NOT EXISTS player_pairs (steamid TEXT PRIMARY KEY,name TEXT DEFAULT '')")
            self.__db_cursor.execute("CREATE TABLE IF NOT EXISTS player_ranks (steamid TEXT PRIMARY KEY,achievementcount INTEGER DEFAULT '0')")

            #Call base class
            AchievementIO.load(self)

        def save(self):
            """
            Saves achievement data to Sqlite database
            """
            self.__db_connection.commit()
            
            #Call base class
            AchievementIO.save(self)
        
        def close(self):
            """
            Closes the Sqlite database cursor
            """
            self.__db_cursor.close()
            
            #Call base class
            AchievementIO.close(self)

        def onRoundEnd(self):
            """
            Is called when a round ends
            """
            for fn in self.__roundEndBuffer:
                if len(fn[1]) == 4:
                    AchievementIOSqlite.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], True)
                elif len(fn[1]) == 6:
                    AchievementIOSqlite.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], fn[1][4], fn[1][5], True)
            
            self.__roundEndBuffer = []

        def getPlayerProgress(self, groupID, achievementID, steamID):
            """
            Retrieves the value of the progress for a particular player for a particular achievement
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerProgress(self, groupID, achievementID, steamID)
            if not cachedValue == None:
                return cachedValue

            self.__db_cursor.execute('SELECT progress FROM achievements_progress WHERE steamid = "%s" AND groupid = "%s" AND achievementID = "%s"' % (steamID, groupID, achievementID))
            for row in self.__db_cursor.fetchone():
                return int(row[0])
            return 0
                    
        def setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly = False, onRoundEnd = False):
            """
            Sets the value of the progress for a particular player for a particular achievement
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerProgress", (self, groupID, achievementID, steamID, progress, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    self.__db_cursor.execute('SELECT progress FROM achievements_progress WHERE steamid = "%s" AND groupid = "%s" AND achievementID = "%s"' % (steamID, groupID, achievementID))
                    exists = False
                    for row in self.__db_cursor.fetchone():
                        exists = True
                    
                    if exists:
                        self.__db_cursor.execute('UPDATE achievements_progress SET progress="%d" WHERE steamid="%s" AND groupid="%s" AND achievementid="%s"' % (int(progress), steamID, groupID, achievementID))
                    else:
                        self.__db_cursor.execute('INSERT INTO achievements_progress (steamid, groupid, achievementid, progress) VALUES ("%s", "%s", "%s", "%d")' % (steamID, groupID, achievementID, int(progress)))                    
                
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly)
                
        def getPlayerName(self, steamID):
            """
            Retrieves the value of the name for a particular player
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerName(self, steamID)
            if not cachedValue == None:
                return cachedValue

            self.__db_cursor.execute('SELECT name FROM player_pairs WHERE steamid="%s"' % steamID)
            for row in self.__db_cursor.fetchone():
                return row[0]
            return ""
        
        def setPlayerName(self, steamID, name, cacheOnly = False, onRoundEnd = False):
            """
            Sets the value of the name for a particular player
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerName", (self, steamID, name, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    self.__db_cursor.execute('SELECT name FROM player_pairs WHERE steamid = "%s"' % steamID)
                    exists = False
                    for row in self.__db_cursor.fetchone():
                        exists = True
                    
                    if exists:
                        self.__db_cursor.execute('UPDATE player_pairs SET name="%s" WHERE steamid="%s"' % (name, steamID))
                    else:
                        self.__db_cursor.execute('INSERT INTO player_pairs (steamid, name) VALUES ("%s", "%s")' % (steamID, name))
              
            #Call base class  
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerName(self, steamID, name, cacheOnly)            
            
        def getPlayerAchievementCount(self, steamID):
            """
            Retrieves the value of the achievement count for a particular player
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerAchievementCount(self, steamID)
            if not cachedValue == None:
                return cachedValue

            self.__db_cursor.execute('SELECT achievementcount FROM player_ranks WHERE steamid="%s"' % steamID)
            for row in self.__db_cursor.fetchone():
                return int(row[0])
            return 0
                   
        def setPlayerAchievementCount(self, steamID, count, cacheOnly = False, onRoundEnd = False):
            """
            Sets the value of the achievment count for a particular player
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerAchievementCount", (self, steamID, count, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    self.__db_cursor.execute('SELECT achievementcount FROM player_ranks WHERE steamid = "%s"' % steamID)
                    exists = False
                    for row in self.__db_cursor.fetchone():
                        exists = True
                    
                    if exists:
                        self.__db_cursor.execute('UPDATE player_ranks SET achievementcount="%d" WHERE steamid="%s"' % (int(count), steamID))
                    else:
                        self.__db_cursor.execute('INSERT INTO player_pairs (steamid, achievementcount) VALUES ("%s", "%d")' % (steamID, int(name)))
            
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerAchievementCount(self, steamID, count, cacheOnly)            
except:
    es.dbgmsg(self.debugLevel, "[MYOA] Error: Could not load sqlite3 module. Cannot use sqlite as a storage method.")
# =======================================================

# == ACHIEVEMENTIOMYSQL CLASS ===========================
try:
    import MySQLdb
    
    class AchievementIOMySQL(AchievementIO):
        def __init__(self, databaseName, host, user, password, onlyRoundEnd = False):
            self.databaseName = databaseName
            self.host = host
            self.user = user
            self.password = password            
            self.onlyRoundEnd = onlyRoundEnd
            self.__roundEndBuffer = []
            self.load()
            
        def load(self):
            """
            Ensures tables are correctly formatted ready for use
            """        
            #Call base class
            AchievementIO.load(self)
                
        def save(self):
            """
            Does nothing special
            """
            #Call base class
            AchievementIO.save(self)
        
        def close(self):
            """
            Does nothing special
            """
            
            #Call base class
            AchievementIO.close(self)
                    
        def setAchievementGroupProperties(self, groupID, title, description):
            """
            Sets the achievement group properties in the database (only for web-based databases)
            """
            db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
            db_cursor = db_connection.cursor()
            
            _groupID = db_connection.escape_string(groupID)
            _title = db_connection.escape_string(title)
            _description = db_connection.escape_string(description) 
            
            db_cursor.execute('SELECT * FROM achievement_groups WHERE stringid="%s"' % _groupID)
            
            if db_cursor.rowcount > 0:
                db_cursor.execute('UPDATE achievement_groups SET title="%s", description="%s" WHERE stringid="%s"' % (_title, _description, _groupID))
            else:
                db_cursor.execute('INSERT INTO achievement_groups (stringid, title, description) VALUES ("%s", "%s", "%s")' % (_groupID, _title, _description))
            
            db_cursor.close()
            db_connection.commit()
            db_connection.close()
    
            #Call base class    
            AchievementIO.setAchievementGroupProperties(self, groupID, title, description)            
        
        def setAchievementProperties(self, groupID, achievementID, title, description, progress, imgPath):
            """
            Sets the achievement propeties in the database (only for web-based databases)
            """
            db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
            db_cursor = db_connection.cursor()
            
            _groupID = db_connection.escape_string(groupID)
            _achievementID = db_connection.escape_string(achievementID)
            _title = db_connection.escape_string(title)
            _description = db_connection.escape_string(description)
            _progress = int(progress)
    
            db_cursor.execute('SELECT * FROM achievements WHERE stringid="%s" AND groupID="%s"' % (_achievementID, _groupID))
            
            if db_cursor.rowcount > 0:
                if imgPath == None:
                    db_cursor.execute('UPDATE achievements SET title="%s", description="%s", progress="%d" WHERE stringid="%s" AND groupID="%s"' % (_title, _description, _progress, _achievementID, _groupID))
                else:
                    db_cursor.execute('UPDATE achievements SET title="%s", description="%s", progress="%d", image="%s" WHERE stringid="%s" AND groupID="%s"' % (_title, _description, _progress, imgPath, _achievementID, _groupID))
            else:
                if imgPath == None:
                    db_cursor.execute('INSERT INTO achievements (groupid, stringid, title, description, progress) VALUES ("%s", "%s", "%s", "%s", "%d")' % (_groupID, _achievementID, _title, _description, _progress))
                else:
                    db_cursor.execute('INSERT INTO achievements (groupid, stringid, title, description, progress, image) VALUES ("%s", "%s", "%s", "%s", "%d", "%s")' % (_groupID, _achievementID, _title, _description, _progress, imgPath))
                
            
            db_cursor.close()
            db_connection.commit()
            db_connection.close()
            
            #Call base class
            AchievementIO.setAchievementProperties(self, groupID, achievementID, title, description, progress, imgPath)            
    
        def onPlayerConnect(self, steamID):
            """
            Is called when a player is connected
            Gets all of the achievement_progress rows belonging to a player
            """
            if not AchievementIO.isPlayerCached(self, steamID):
                db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
                db_cursor = db_connection.cursor()
                
                _steamID = db_connection.escape_string(steamID)
                
                db_cursor.execute('SELECT * FROM achievements_progress WHERE steamid="%s"' % _steamID)
                rows = db_cursor.fetchall()
                for row in rows:
                    AchievementIO.setPlayerProgress(self, row[1], row[2], row[0], int(row[3]), True)
                
                db_cursor.execute('SELECT * FROM player_pairs WHERE steamid = "%s" LIMIT 1' % _steamID)
                rows = db_cursor.fetchall()
                for row in rows:
                    AchievementIO.setPlayerName(self, row[1], row[2], True)
        
                db_cursor.execute('SELECT * FROM player_ranks WHERE steamid = "%s" LIMIT 1' % _steamID)
                rows = db_cursor.fetchall()
                for row in rows:
                    AchievementIO.setPlayerAchievementCount(self, row[0], row[1], True)
                
                db_cursor.close()
                db_connection.commit()
                db_connection.close()
                
            #Call base class
            AchievementIO.onPlayerConnect(self, steamID)
    
        def onRoundEnd(self):
            """
            Is called when a round ends
            """
            if len(self.__roundEndBuffer) > 0:
                db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
                db_cursor = db_connection.cursor()

                for fn in self.__roundEndBuffer:
                    if len(fn[1]) == 4:
                        AchievementIOMySQL.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], True, db_cursor)
                    elif len(fn[1]) == 5:
                        AchievementIOMySQL.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], fn[1][4], True, db_cursor)
                    elif len(fn[1]) == 6:
                        AchievementIOMySQL.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], fn[1][4], fn[1][5], True, db_cursor)
                
                db_cursor.close()
                db_connection.commit()
                db_connection.close()
            
            self.__roundEndBuffer = []
                
        def getPlayerProgress(self, groupID, achievementID, steamID):
            """
            Retrieves the value of the progress for a particular player for a particular achievement
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerProgress(self, groupID, achievementID, steamID)
            if not cachedValue == None:
                return cachedValue
            return 0
                    
        def setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly = False, onRoundEnd = False, db_cursor = None):
            """
            Sets the value of the progress for a particular player for a particular achievement
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerProgress", (self, groupID, achievementID, steamID, progress, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    inherit = (not db_cursor == None)
                    if not inherit: 
                        db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
                        db_cursor = db_connection.cursor()
                    
                    _groupID = db_connection.escape_string(groupID)
                    _achievementID = db_connection.escape_string(achievementID)
                    _steamID = db_connection.escape_string(steamID)
                    _progress = int(progress)
                    
                    db_cursor.execute('SELECT progress FROM achievements_progress WHERE steamid = "%s" AND groupid = "%s" AND achievementID = "%s"' % (_steamID, _groupID, _achievementID))
                    
                    if db_cursor.rowcount > 0:
                        db_cursor.execute('UPDATE achievements_progress SET progress="%d" WHERE steamid="%s" AND groupid="%s" AND achievementid="%s"' % (_progress, _steamID, _groupID, _achievementID))
                    else:
                        db_cursor.execute('INSERT INTO achievements_progress (steamid, groupid, achievementid, progress) VALUES ("%s", "%s", "%s", "%d")' % (_steamID, _groupID, _achievementID, _progress))
    
                    if not inherit:
                        db_cursor.close()
                        db_connection.commit()
                        db_connection.close()
                
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly)
                
        def getPlayerName(self, steamID):
            """
            Retrieves the value of the name for a particular player
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerName(self, steamID)
            if not cachedValue == None:
                return cachedValue
            return ""
        
        def setPlayerName(self, steamID, name, cacheOnly = False, onRoundEnd = False, db_cursor = None):
            """
            Sets the value of the name for a particular player
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerName", (self, steamID, name, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    inherit = (not db_cursor == None)
                    if not inherit: 
                        db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
                        db_cursor = db_connection.cursor()
    
                    _steamID = db_connection.escape_string(steamID)
                    _name = db_connection.escape_string(name)
                    
                    db_cursor.execute('SELECT name FROM player_pairs WHERE steamid = "%s"' % _steamID)
                    
                    if db_cursor.rowcount > 0:
                        db_cursor.execute('UPDATE player_pairs SET name="%s" WHERE steamid="%s"' % (_name, _steamID))
                    else:
                        db_cursor.execute('INSERT INTO player_pairs (steamid, name) VALUES ("%s", "%s")' % (_steamID, _name))
                    
                    if not inherit:
                        db_cursor.close()
                        db_connection.commit()
                        db_connection.close()
              
            #Call base class  
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerName(self, steamID, name, cacheOnly)            
            
        def getPlayerAchievementCount(self, steamID):
            """
            Retrieves the value of the achievement count for a particular player
            """
            #Call base class to see if there is a cached value - if not, continue on
            cachedValue = AchievementIO.getPlayerAchievementCount(self, steamID)
            if not cachedValue == None:
                return cachedValue
            return 0
                   
        def setPlayerAchievementCount(self, steamID, count, cacheOnly = False, onRoundEnd = False, db_cursor = None):
            """
            Sets the value of the achievment count for a particular player
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerAchievementCount", (self, steamID, count, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    inherit = (not db_cursor == None)
                    if not inherit: 
                        db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
                        db_cursor = db_connection.cursor()
    
                    _steamID = db_connection.escape_string(steamID)
                    _count = int(count)
    
                    db_cursor.execute('SELECT achievementcount FROM player_ranks WHERE steamid = "%s"' % _steamID)
                    
                    if db_cursor.rowcount > 0:
                        db_cursor.execute('UPDATE player_ranks SET achievementcount="%d" WHERE steamid="%s"' % (_count, _steamID))
                    else:
                        db_cursor.execute('INSERT INTO player_ranks (steamid, achievementcount) VALUES ("%s", "%d")' % (_steamID, _count))
    
                    if not inherit:
                        db_cursor.close()
                        db_connection.commit()
                        db_connection.close()
            
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerAchievementCount(self, steamID, count, cacheOnly)            
    
        def onPlayerAchievementEvent(self, steamID, groupID, achievementID, event, onRoundEnd = False, db_cursor = None):
            """
            Is called when a player triggers an achievement event (achieved/un-achieved/reset)
            """
            if self.onlyRoundEnd and not onRoundEnd:
                appendage = ["onPlayerAchievementEvent", (self, steamID, groupID, achievementID, event)]
                self.__roundEndBuffer.append(appendage)
            else:
                inherit = (not db_cursor == None)
                if not inherit: 
                    db_connection = MySQLdb.connect(host = str(self.host), user = str(self.user), passwd = str(self.password), db = str(self.databaseName))
                    db_cursor = db_connection.cursor()
                
                _steamID = db_connection.escape_string(steamID)
                _groupID = db_connection.escape_string(groupID)
                _achievementID = db_connection.escape_string(achievementID)
                _event = db_connection.escape_string(event)
                    
                db_cursor.execute('INSERT INTO achievement_events (date, steamid, groupid, achievementid, event) VALUES (NOW(), "%s", "%s", "%s", "%s")' % (_steamID, _groupID, _achievementID, _event))
                
                if not inherit:
                    db_cursor.close()
                    db_connection.commit()
                    db_connection.close()            
            
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.onPlayerAchievementEvent(self, steamID, groupID, achievementID, event)
except:
    es.dbgmsg(self.debugLevel, "[MYOA] Error: Could not load MySQLdb module. Cannot use MySQL as a storage method.")
# =======================================================

# == ACHIEVEMENTIOMYSQLPHP CLASS ===========================
try:
    import subprocess
    import os
    
    class AchievementIOMySQLPHP(AchievementIO):
        def __init__(self, url, secretKey, onlyRoundEnd = False):
            self.__onPath = True
            self.url = url
            self.secretKey = secretKey
            self.onlyRoundEnd = onlyRoundEnd
            self.__roundEndBuffer = []
            self.load()
            
        def load(self):
            """
            Does nothing, since all loads/saves will be done via PHP
            """
            #Call base class
            AchievementIO.load(self)
            
        def save(self):
            """
            Does nothing, since all loads/saves will be done via PHP
            """
            #Call base class
            AchievementIO.save(self)
        
        def close(self):
            """
            Does nothing, since all loads/saves will be done via PHP
            """
            #Call base class
            AchievementIO.close(self)
        
        def __tryWGet(self, fullUrl):
            """
            "Private" method, which trys to run wget in the background with the url passed
            """
            if self.__onPath:
                try:
                    os.system("wget -q -b %s" % fullUrl)
                except:
                    self.__onPath = False
                    self.__tryWGet(fullUrl)
            else:
                try:
                    wgetPath = es.getAddonPath("myo_achievement") + "/wget.exe"                        
                    os.system("%s -q -b %s" % (wgetPath, fullUrl))
                except:
                    es.dbgmsg(self.debugLevel, "Cannot run wget. Ensure that either wget is enclosed in the myo_achievement addon folder, or that it can be accessed in a folder on the PATH variable.")            
        
        def setAchievementGroupProperties(self, groupID, title, description):
            """
            Sets the achievement group properties in the database (only for web-based databases)
            """
            fullUrl = "%s?mode=setgroup&sk=%s&gid=%s&ti=%s&de=%s" % (self.url, self.secretKey, groupID, title, description)
            self.__tryWGet(fullUrl.replace(" ", "%20"))
            
            #Call base class    
            AchievementIO.setAchievementGroupProperties(self, groupID, title, description)            
        
        def setAchievementProperties(self, groupID, achievementID, title, description, progress, imgPath):
            """
            Sets the achievement propeties in the database (only for web-based databases)
            """
            fullUrl = "%s?mode=setachievement&sk=%s&gid=%s&aid=%s&ti=%s&de=%s&val=%d" % (self.url, self.secretKey, groupID, achievementID, title, description, int(progress))
            if not imgPath == None:
                fullUrl = "%s&img=%s" % (fullUrl, imgPath)
            self.__tryWGet(fullUrl.replace(" ", "%20"))
            
            #Call base class
            AchievementIO.setAchievementProperties(self, groupID, achievementID, title, description, progress, imgPath)            

        def onPlayerConnect(self, steamID):
            """
            Is called when a player is connected
            Asks a PHP file for RCON feedback for a player
            """
            if not AchievementIO.isPlayerCached(self, steamID):
                fullUrl = "%s?mode=get&sk=%s&id=%s&ip=%s&port=%s" % (self.url, self.secretKey, steamID, es.ServerVar("ip"), es.ServerVar("hostport"))
                self.__tryWGet(fullUrl.replace(" ", "%20"))
                
            #Call base class
            AchievementIO.onPlayerConnect(self, steamID)
        
        def onRoundEnd(self):
            """
            Is called when a round ends
            """
            for fn in self.__roundEndBuffer:
                if len(fn[1]) == 4:
                    AchievementIOMySQLPHP.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], True)
                elif len(fn[1]) == 5:
                    AchievementIOMySQLPHP.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], fn[1][4], True)
                elif len(fn[1]) == 6:
                    AchievementIOMySQLPHP.__dict__[fn[0]](fn[1][0], fn[1][1], fn[1][2], fn[1][3], fn[1][4], fn[1][5], True)
            
            self.__roundEndBuffer = []

        def getPlayerProgress(self, groupID, achievementID, steamID):
            """
            Retrieves the value of the progress for a particular player for a particular achievement
            """
            #Call base class to see if there is a cached value - if not, return 0
            cachedValue = AchievementIO.getPlayerProgress(self, groupID, achievementID, steamID)
            if not cachedValue == None:
                return cachedValue
            return 0
                    
        def setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly = False, onlyRoundEnd = False):
            """
            Sets the value of the progress for a particular player for a particular achievement
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerProgress", (self, groupID, achievementID, steamID, progress, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    fullUrl = "%s?mode=setprogress&sk=%s&gid=%s&aid=%s&id=%s&val=%s" % (self.url, self.secretKey, groupID, achievementID, steamID, progress)
                    self.__tryWGet(fullUrl.replace(" ", "%20"))
            
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerProgress(self, groupID, achievementID, steamID, progress, cacheOnly)

        def getPlayerName(self, steamID):
            """
            Retrieves the value of the name for a particular player
            """
            #Call base class to see if there is a cached value - if not, return ""
            cachedValue = AchievementIO.getPlayerName(self, steamID)
            if not cachedValue == None:
                return cachedValue
            return ""
        
        def setPlayerName(self, steamID, name, cacheOnly = False, onlyRoundEnd = False):
            """
            Sets the value of the name for a particular player
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerName", (self, steamID, name, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    fullUrl = "%s?mode=setname&sk=%s&id=%s&val=%s" % (self.url, self.secretKey, steamID, name)
                    self.__tryWGet(fullUrl.replace(" ", "%20"))

            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerName(self, steamID, name, cacheOnly)
            
        def getPlayerAchievementCount(self, steamID):
            """
            Retrieves the value of the achievement count for a particular player
            """
            #Call base class to see if there is a cached value - if not, return 0
            cachedValue = AchievementIO.getPlayerAchievementCount(self, steamID)
            if not cachedValue == None:
                return cachedValue
            return 0
        
        def setPlayerAchievementCount(self, steamID, count, cacheOnly = False, onlyRoundEnd = False):
            """
            Sets the value of the achievement count for a particular player
            """
            if not cacheOnly:
                if self.onlyRoundEnd and not onRoundEnd:
                    appendage = ["setPlayerAchievementCount", (self, steamID, count, cacheOnly)]
                    self.__roundEndBuffer.append(appendage)
                else:
                    fullUrl = "%s?mode=setcount&sk=%s&id=%s&val=%s" % (self.url, self.secretKey, steamID, count)
                    self.__tryWGet(fullUrl.replace(" ", "%20"))

            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:
                AchievementIO.setPlayerAchievementCount(self, steamID, count, cacheOnly)
        
        def onPlayerAchievementEvent(self, steamID, groupID, achievementID, event, onlyRoundEnd = False):
            """
            Is called when a player triggers an achievement event (achieved/un-achieved/reset)
            """            
            if self.onlyRoundEnd and not onRoundEnd:
                appendage = ["onPlayerAchievementEvent", (self, steamID, groupID, achievementID, event)]
                self.__roundEndBuffer.append(appendage)
            else:
                fullUrl = "%s?mode=achievementevent&sk=%s&id=%s&gid=%s&aid=%s&val=%s" % (self.url, self.secretKey, steamID, groupID, achievementID, event)
                self.__tryWGet(fullUrl.replace(" ", "%20"))
            
            #Call base class
            if not self.onlyRoundEnd or not onRoundEnd:            
                AchievementIO.onPlayerAchievementEvent(self, steamID, groupID, achievementID, event)
except:
    es.dbgmsg(self.debugLevel, "[MYOA] Error: Could not load subprocess module. Cannot use MySQL via PHP as a storage method.")
# =======================================================